import java.util.*;
import java.io.*;

/**
 * This class represents a finite automaton in both
 * non-determinstic and determinstic variants.
 *
 */
public class Automaton {
	// Tabular representation of automaton transition matrix
	private Set[][] automaton;
	private int numStates, numLabels;
	// Automaton consists of states and labels, transitions are managed by states
	private Set states, labels;
	private State currentState;
	// Utility data structure to map unique names to corresponding instance
	private Map nameToState, nameToLabel;
        // Utility data structure to map integer id to states and labels
	private List stateList, labelList;

	// Teilaufgabe a) ---------------------------------------------------------

	public Automaton() {
		numStates = 0; numLabels = 0;
		states = new HashSet(); labels = new HashSet();
		nameToState = new HashMap(); nameToLabel = new HashMap();
		stateList = new Vector(); labelList = new Vector();
	}

	// Add a new state to the automaton, possibly as start or terminal state
	public void addState(String name, boolean isStart, boolean isTerminal) {
		State s = (State) nameToState.get(name);
		if (s == null) {
			s = new State(name, numStates);
			s.setStart(isStart); s.setTerminal(isTerminal);
			states.add(s);
                        stateList.add(s);
			nameToState.put(name, s);
			numStates++;
		}
	}

	// Add a new transition to the automaton between possibly existing states
	public void addTransition(String from, String to, String label) {
		State origin = (State) nameToState.get(from);
		State target = (State) nameToState.get(to);
		if (origin == null) {
			addState(from, false, false);
			origin = (State) nameToState.get(from);
		}
		if (target == null) {
			addState(to, false, false);
			target = (State) nameToState.get(to);
		}
		// Create label if necessary
		Label l = (Label) nameToLabel.get(label);
		if (l == null) {
			l = new Label(label, numLabels);
			labels.add(l);
                        labelList.add(l);
			nameToLabel.put(label, l);
			numLabels++;
		}
		// Create transition itself
		Transition t = new Transition(origin, target, l);
		origin.getOutgoingTransitions().add(t);
		target.getIncomingTransitions().add(t);
	}

        // Retrieve current state of the automaton
        public State getCurrentState() {
                return currentState;
        }

        // Set current state of the automaton
        public void setCurrentState(State s) {
                currentState = s;
        }

	// Initialize the automaton, called before execution
	public void init() {
		initTransitionMatrix();
		// Choose an initial state
		Iterator sit = states.iterator();
		while (sit.hasNext()) {
			State s = (State) sit.next();
			if (s.isStart()) {
				currentState = s;
				break;
			}
		}
		if (currentState == null)
			System.out.println("*** No initial state!");
	}

	/**
         *  Execute a single step of the automaton by providing a language symbol,
         *  not changing its internal state (call setCurrentState() later to do so)
	 *  Returns a set of states which is either
	 *    null, if no valid symbol has been provided,
	 *    empty, if no valid transition exists (i.e. the word is not accepted),
	 *    a single state, if the automaton is deterministic for given symbol,
	 *    a number of states, if the automaton is nondeterministic for given symbol.
	 * */
	public Set step(String symbol) {
		Label label = (Label) nameToLabel.get(symbol);
		if (label == null) {
			System.out.println("*** Token \"" + symbol + "\" not recognized!");
			return null;
		}
		Set result = automaton[currentState.getId()][label.getId()];
		if (result.isEmpty()) {
			System.out.println("\n*** Token \"" + symbol + "\" not valid in state \""
			+ currentState.getName() +"\"!");
			System.out.print("*** Expected symbol(s): ");
			for (int j = 0; j < numLabels; j++) {
				if (automaton[currentState.getId()][j].size() > 0)
					System.out.print(labelList.get(j) + " ");
			}
			System.out.print("\n");
			return result;
		}
		return result;
	}

	// Teilaufgabe b) ---------------------------------------------------------

	// Construct automaton from given regular grammar with productions of form
	// <nonterminal> -> <epsilon>|terminal|terminal <nonterminal>
	public void readGrammar(String fileName) {
		// Initialize private fields
		numStates = 0; numLabels = 0; currentState = null;
		nameToState = new HashMap(); nameToLabel = new HashMap();
		states = new HashSet(); labels = new HashSet();

		// Iterate all productions in given grammar (no error checking)
		String buffer = new String();
		String lhs = new String(), rhs1 = new String(), rhs2 = new String();
		try {
			BufferedReader reader = new BufferedReader(new FileReader(fileName));
			while (reader.ready()) {
				// Separate productions into left- and right-hand side parts
				buffer = reader.readLine();
				StringTokenizer tokenizer = new StringTokenizer(buffer);
				lhs = tokenizer.nextToken();
				tokenizer.nextToken();
				rhs1 = tokenizer.nextToken();
				if (tokenizer.hasMoreTokens())
					rhs2 = tokenizer.nextToken();
				else
					rhs2 = "";

				// Process originating state i.e. nonterminal on left-hand side
				addState(lhs, isStart(lhs), false);
				if (isEpsilon(rhs1)) {
					// Production: '<nonterminal> -> 'epsilon>'
					State origin = (State) nameToState.get(lhs);
					origin.setTerminal(true);
				} else {
					// Process target state if necessary
					if (rhs2.length() > 0) {
						// Production: '<nonterminal> -> terminal <nonterminal>'
						addState(rhs2, false, false);
						addTransition(lhs, rhs2, rhs1);
					} else {
						// Production: '<nonterminal> -> terminal'
						// Add dedicated final state
						addState("<F>", false, true);
						addTransition(lhs, "<F>", rhs1);
					}
				}
			}
			reader.close();
		}
		catch (IOException ex) {
			ex.printStackTrace();
			System.exit(1);
		}
	}

	// Initialize transition matrix after change to automaton structure
	private void initTransitionMatrix() {
		automaton = new Set [numStates][numLabels];
		for (int i = 0; i < numStates; i++) {
			for (int j = 0; j < numLabels; j++) {
				automaton[i][j] = new HashSet();
			}
		}
		Iterator sit = states.iterator();
		while (sit.hasNext()) {
			State s = (State) sit.next();
			Iterator tit = s.getOutgoingTransitions().iterator();
			while (tit.hasNext()) {
				Transition t = (Transition) tit.next();
				automaton[s.getId()][t.getLabel().getId()].add(t.getTarget());
			}
		}
	}

	// Detect a number of special tokens
	private boolean isEpsilon(String token) {
		if (token.equals("<epsilon>"))
			return true;
		return false;
	}
	private boolean isStart(String token) {
		if (token.equals("<S>"))
			return true;
		return false;
	}

	// Formatted output of automaton
	public String toString() {
		String result = new String();
		result += "States: ";
		Iterator sit = states.iterator();
		while (sit.hasNext()) {
			result += (State) sit.next();
			if (sit.hasNext()) result += ", ";
		}
		result += "\nLabels: ";
		Iterator lit = labels.iterator();
		while (lit.hasNext()) {
			result += (Label) lit.next();
			if (lit.hasNext()) result += ", ";
		}
		result += "\nTransitions:\n";
		sit = states.iterator();
		while (sit.hasNext()) {
			State s = (State) sit.next();
			Iterator tit = s.getOutgoingTransitions().iterator();
			while (tit.hasNext()) {
				Transition t = (Transition) tit.next();
				result += "\t" + s + " -" + t + "-> " + t.getTarget() +"\n";
			}
		}
		result += "\n";
		return result;
	}

	// Dump the transition matrix - just some fancy formatting code
	public String matrixToString() {
		String result = new String();
		Iterator sit = states.iterator();
		result += "States: ";
		while (sit.hasNext()) {
			State s = (State) sit.next();
			result += s.getName() + ":" + s.getId();
			if (sit.hasNext()) result += ", "; else result += "\n";
		}
		Iterator lit = labels.iterator();
		result += "Labels: ";
		while (lit.hasNext()) {
			Label l = (Label) lit.next();
			result += l.getName() + ":" + l.getId();
			if (lit.hasNext()) result += ", "; else result += "\n";
		}
		result += "\t|";
		for (int j = 0; j < numLabels; j++) {
			result += j + "\t\t|";
		}
		result += "\n";
		for (int i = 0; i < numStates; i++) {
			result += i + "\t|";
			for (int j = 0; j < numLabels; j++) {
				Iterator targets = automaton[i][j].iterator();
				while (targets.hasNext()) {
					result += ((State) targets.next()).getId();
					if (targets.hasNext()) result += ",";
				}
				if (automaton[i][j].size() > 1) result += "\t|";
				else result += "\t\t|";
			}
			result += "\n";
		}
		return result;
	}
}
